(function(window) {
	var Piece = function(canvas, config)
	{
		this.initialize(canvas, config);
	}
	var p = Piece.prototype = new BasePiece();
	//
	p.initialize = function(canvas, config)
	{
		BasePiece.prototype.initialize.call(this, canvas, config);
		this.initInteraction();
		if (this.config.soundOn) this.initSound();
	}

	p.initSound = function()
	{
		var sounds = [];
		for (var i=0;i<this.config.sounds.length;i++) sounds.push("sounds/circle-bells-"+this.config.sounds[i]);
		sounds = ArrayUtil.shuffle(sounds);
		SoundManager.init(sounds);
		SoundManager.max = isMobile ? this.config.soundsMaxSimultaneousMobile : this.config.soundsMaxSimultaneous;
		this.soundStarted = !isMobile;
		this.soundId = null;
		this.lastSoundAt = 0;
	}
	p.randomizeSound = function()
	{
		this.soundId = RandomUtil.pick(SoundManager.loaded, [this.soundId]);
	}


	p.onKeyUp = function(e)
	{
		BasePiece.prototype.onKeyUp.call(this, e);
		if (!this.config.debug) return;
		var c = String.fromCharCode(e.which);
		if (c=="R") this.reset();
	}
	p.initInteraction = function()
	{
		var that = this;
		var type = isMobile?"stagemousedown":"stagemouseup";
		this.stage.addEventListener(type, this.handleClick.bind(this));
	}

	p.handleClick = function(e)
	{
		//debug:
		if (this.config.toggleOn)
		{
			this.toggleBlur();
		}
		//start sound on user interaction (if not already started)
		if (!this.soundStarted)
		{
			this.playSound();
			this.soundStarted = true;
		}
	}

	p.blurred = false;
	p.toggleBlur = function()
	{
		this.blurred = !this.blurred;
		var style = "";
		if (this.blurred)
		{
			var b = "filter:blur(" + this.config.blur + "px);"
			style = "-webkit-"+b+" -moz-"+b+" -ms-"+b+" -o-"+b+" "+b+" transform: translateZ(0);";
		}
		this.canvas.setAttribute("style",style);
		this.canvas.style.width = Math.round(this.width/this.dpr)+"px";
		this.canvas.style.height = Math.round(this.height/this.dpr)+"px";
	}

	p.setSize = function(w,h,dpr)
	{
		this.dpr = dpr;
		this.width = Math.floor(w*dpr);
		this.height = Math.floor(h*dpr);
		this.area = this.width*this.height;
		this.diam = Math.sqrt(w*w+h*h)*dpr;
		log("setSize",w,this.width);
		var r = this.radius = Math.sqrt(this.area * this.config.ballSize / Math.PI);
		this.distSquare = r*r*4;
		this.bounds = [r,r,this.width-r, this.height-r];
		if (this.tickLast) this.reset();
	}

	p.start = function()
	{
		BasePiece.prototype.start.apply(this);
		log("start",this.width);
		this.balls = [];
		if (this.width) this.reset();
		this.toggleBlur();
	}

	p.reset = function()
	{
		this.balls.length = 0;
		this.stage.removeAllChildren();
		//add balls
		var r = this.radius, w = this.width-2*r, h = this.height-2*r;
		var dsq = this.distSquare;
		for (var i=0;i<3;i++)
		{
			var n = this.balls.length;
			for (var a=0;a<199;a++)
			{
				var x = Math.random()*w+r, y =  Math.random()*h+r;
				var ok = true;
				for (var j=0;j<n;j++)
				{
					var b = this.balls[j];
					var dx = b.x-x, dy = b.y-y;
					if (dx*dx+dy*dy < dsq)
					{
						ok = false;
						break;
					}
				}
				if (ok)
				{
					this.addBall(x,y, this.config.colors[i]);
					break;
				}
			}
		}
	}

	p.addBall = function(x,y, color)
	{
		var r = this.radius;
		var s = this.speed = this.config.ballSpeed * this.diam;
		var b = new Shape();
		b.x = x;
		b.y = y;
		var a = Math.random()*Math.PI*2;
		b.vx = Math.cos(a) * s;
		b.vy = Math.sin(a) * s;
		//draw
		//b.graphics.s(color).dc(0,0,r);
		b.graphics.f(color).dc(0,0,r);
		b.cache(-r,-r,2*r,2*r);
		//add/store
		this.stage.addChild(b);
		this.balls.push(b);
	}

	p.onCollision = function()
	{
		//swap balls around
		var props = {x:0,y:0,vx:0,vy:0};
		var n = 3;
		for (var i=0;i<n-1;i++)
		{
			var b0 = this.balls[i];
			var b1 = this.balls[(i+1)%n];
			for (var s in props)
			{
				props[s] = b0[s];
				b0[s] = b1[s];
				b1[s] = props[s];
			}
		}
		//play sound
		this.playSound();
	}

	p.playSound = function()
	{
		if (!this.config.soundOn) return;
		if (Date.now()-this.lastSoundAt>this.config.soundsTimeBetweenMin)
		{
			this.randomizeSound();
			if (this.soundId)
			{
				SoundManager.play(this.soundId);
				this.lastSoundAt = this.tickLast;
			}
		}
	}
	p.normalizeSpeed = function(ball)
	{
		var b = ball;
		var nf = this.speed / Math.sqrt(b.vx*b.vx+b.vy*b.vy);
		b.vx *= nf;
		b.vy *= nf;
	}

	p.update = function()
	{
		var n = this.balls.length;
		var bs = this.bounds, xmin = bs[0], xmax = bs[2], ymin = bs[1], ymax = bs[3];
		var d = this.radius*2, dsq = this.distSquare;
		var sc = this.config.ballSpeedConstant;
		var np = 10;//nr of passes
		var collisionBall = false;
		var collisionViewPort = false;
		for (var p=0;p<np;p++)
		{
			for (var i=0;i<n;i++)
			{
				var b = this.balls[i];
				//collisions
				for (var j=i+1;j<n;j++)
				{
					var b1 = this.balls[j];
					var dx = b1.x-b.x, dy = b1.y-b.y;
					var db = dx*dx+dy*dy;
					if (db<dsq)
					{
						var a = Math.atan2(dy,dx);
						var ax = Math.cos(a)*d - dx;
						var ay = Math.sin(a)*d - dy;
						b.vx -= ax;
						b.vy -= ay;
						b1.vx += ax;
						b1.vy += ay;
						if (sc)
						{
							//normalize speeds
							this.normalizeSpeed(b);
							this.normalizeSpeed(b1);
						}
						//
						collisionBall = true;
					}
				}
				//move
				b.x += b.vx/np;
				b.y += b.vy/np;
				if (b.x>xmax)
				{
					b.x = xmax;
					b.vx *= -1;
					collisionViewPort = true;
				}
				else if (b.x < xmin) {
					b.x = xmin;
					b.vx *= -1;
					collisionViewPort = true;
				}
				if (b.y>ymax)
				{
					b.y = ymax;
					b.vy *= -1;
					collisionViewPort = true;
				}
				else if (b.y < ymin) {
					b.y = ymin;
					b.vy *= -1;
					collisionViewPort = true;
				}
			}
		}
		if (this.config.changeColorsOnViewportCollision && collisionViewPort ||
			this.config.changeColorsOnBallCollision && collisionBall)
		{
			this.onCollision();
		}
		return true;
	}


	window.Piece = Piece;
}(window));

